/*
-------------------------------------------------------------------------------
File Name : readpng.c

Purpose : Implementation of read_png_info, read_png_gray and write_png_gray.

Creation Data : 2010-01-13

Last Modified : ti. 19. jan. 2010 kl. 14.15 +0100

Created By : Martin Ertsås
-------------------------------------------------------------------------------
*/

#include "readpng.h"

/*
 * Get width, height and depth of a .PNG file. 
 */
int read_png_info(char *filename, int *width, int *height, int *bit_depth)
{
    //Check first eight bytes to see if it's a PNG file.
    fp = fopen(filename, "rb");
    if(!fp) {
        return 1;
    }

    png_byte header[HEADER_SIZE];
    int is_png;

    fread(header, 1, HEADER_SIZE, fp);
    is_png = !png_sig_cmp(header, 0, HEADER_SIZE);

    if(!is_png) { //Not a PNG file.
        fclose(fp);
        return 2;
    }

    //Make the info and read structs.
    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if(!png_ptr) {
        perror("[read_png_info] Error making read struct\n");
        fclose(fp);
        return 4;
    }

    info_ptr = png_create_info_struct(png_ptr);
    if(!info_ptr) {
        perror("[read_png_info] Error making info struct\n");
        png_destroy_read_struct(&png_ptr, NULL, NULL);
        fclose(fp);
        return 4;
    }

    stop_info = png_create_info_struct(png_ptr);
    if(!stop_info) {
        perror("[read_png_info] Error making stop info struct\n");
        png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
        fclose(fp);
        return 4;
    }

    if(setjmp(png_jmpbuf(png_ptr))) {
        perror("[read_png_info] Error initiating IO\n");
        png_destroy_read_struct(&png_ptr, &info_ptr, &stop_info);
        fclose(fp);
        return 4;
    }
    png_init_io(png_ptr, fp);

    if(setjmp(png_jmpbuf(png_ptr))) {
        perror("[read_png_info] Error skipping eight bytes\n");
        png_destroy_read_struct(&png_ptr, &info_ptr, &stop_info);
        fclose(fp);
        return 4;
    }
    png_set_sig_bytes(png_ptr, HEADER_SIZE);

    //Read header
    if(setjmp(png_jmpbuf(png_ptr))) {
        perror("[read_png_info] Error reading header\n");
        png_destroy_read_struct(&png_ptr, &info_ptr, &stop_info);
        fclose(fp);
        return 4;
    }
    png_read_info(png_ptr, info_ptr);

    //Check if the file can be read.
    png_uint_32 _width, _height;
    int depth, interlace, compression, filter;

    if(setjmp(png_jmpbuf(png_ptr))) {
        perror("[read_png_info] Error getting header info\n");
        png_destroy_read_struct(&png_ptr, &info_ptr, &stop_info);
        fclose(fp);
        return 4;
    }

    png_get_IHDR(png_ptr, info_ptr, &_width, &_height, &depth, &color, 
            &interlace, &compression, &filter);

    if(filter != PNG_FILTER_TYPE_BASE) {
        perror("[read_png_info] Filter type not supported\n");
        png_destroy_read_struct(&png_ptr, &info_ptr, &stop_info);
        fclose(fp);
        return 4;
    }

    if(compression != PNG_COMPRESSION_TYPE_BASE) {
        perror("[read_png_info] Compression type not supported\n");
        png_destroy_read_struct(&png_ptr, &info_ptr, &stop_info);
        fclose(fp);
        return 4;
    }

    if(_width == 0 || _height == 0 || depth == 0) {
        perror("[read_png_info] Size and/or depth of the image can not be zero\n");
        png_destroy_read_struct(&png_ptr, &info_ptr, &stop_info);
        fclose(fp);
        return 4;
    }

    //image can be read safely.
    *width = _width;
    *height = _height;
    *bit_depth = depth;
    return 0;
}

/*
 * Read grayscale image.
 */
int read_png_gray(double **image, int width, int height)
{
    //Check for error in read_png_info.
    int i, j;
    if(!fp || !png_ptr || !info_ptr || !stop_info) {
        return 1;
    }

    if(color != PNG_COLOR_TYPE_GRAY) {
        return 2;
    }

    //Allocate memory.
    uint8_t **_image;
    _image = malloc(height*sizeof(uint8_t*));
    for(i = 0; i < height; i++) {
        _image[i] = malloc(width*sizeof(uint8_t));
    }

    //Read the image.
    //Check for unexpected error.
    if(setjmp(png_jmpbuf(png_ptr))) {
        perror("[read_png_gray] Error reading file\n");
        png_destroy_read_struct(&png_ptr, &info_ptr, &stop_info);
        fclose(fp);
        return 3;
    }

    png_read_image(png_ptr, _image);

    //Copy to the double array and clean up.
    for(i = 0; i < height; i++) {
        for(j = 0; j < width; j++) {
            image[i][j] = (double) _image[i][j];
        }
        free(_image[i]);
    }
    free(_image);

    png_destroy_read_struct(&png_ptr, &info_ptr, &stop_info);
    fclose(fp);
    return 0;
}

/*
 * Read rgb image.
 */
int read_png_rgb(double **red, double **green, double **blue, int width, 
        int height)
{
    int i, j;
    if(!fp || !png_ptr || !info_ptr || !stop_info) {
        return 1;
    }

    if(color != PNG_COLOR_TYPE_RGB) {
        return 2;
    }

    uint8_t **image;
    image = malloc(height*sizeof(uint8_t*));
    for(i = 0; i < height; i++) {
        image[i] = malloc(3*width*sizeof(uint8_t));
    }

    if(setjmp(png_jmpbuf(png_ptr))) {
        perror("[read_png_rgb] Error reading file\n");
        png_destroy_read_struct(&png_ptr, &info_ptr, &stop_info);
        fclose(fp);
        return 3;
    }

    png_read_image(png_ptr, image);

    for(i = 0; i < height; i++) {
        for(j = 0; j < width; j++) {
            red[i][j] = (double) image[i][j*3];
            green[i][j] = (double) image[i][j*3 + 1];
            blue[i][j] = (double) image[i][j*3 + 2];
        }
        free(image[i]);
    }
    free(image);

    png_destroy_read_struct(&png_ptr, &info_ptr, &stop_info);
    fclose(fp);
    return 0;
}

/*
 * Read rgba image.
 */
int read_png_rgba(double **red, double **green, double **blue, double **alpha,
        int width, int height)
{
    int i, j;
    if(!fp || !png_ptr || !info_ptr || !stop_info) {
        return 1;
    }

    if(color != PNG_COLOR_TYPE_RGBA) {
        return 2;
    }

    uint8_t **image;
    image = malloc(height*sizeof(uint8_t*));
    for(i = 0; i < height; i++) {
        image[i] = malloc(4*width*sizeof(uint8_t));
    }

    if(setjmp(png_jmpbuf(png_ptr))) {
        perror("[read_png_rgba] Error reading file\n");
        png_destroy_read_struct(&png_ptr, &info_ptr, &stop_info);
        fclose(fp);
        return 3;
    }

    png_read_image(png_ptr, image);

    for(i = 0; i < height; i++) {
        for(j = 0; j < width; j++) {
            red[i][j] = image[i][j*4];
            green[i][j] = image[i][j*4 + 1];
            blue[i][j] = image[i][j*4 + 2];
            alpha[i][j] = image[i][j*4 + 3];
        }
        free(image[i]);
    }
    free(image);

    png_destroy_read_struct(&png_ptr, &info_ptr, &stop_info);
    fclose(fp);
    return 0;
}

/*
 * Write PNG image.
 */
int write_png_gray(char *filename, double **image, int width, int height, 
        int bit_depth)
{
    fp = fopen(filename, "wb");
    if(!fp) {
        return 1;
    }

    png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if(!png_ptr) {
        perror("[write_png_gray] Error making write struct\n");
        fclose(fp);
        return 2;
    }

    info_ptr = png_create_info_struct(png_ptr);
    if(!info_ptr) {
        perror("[write_png_gray] Error making info struct\n");
        png_destroy_write_struct(&png_ptr, NULL);
        fclose(fp);
        return 2;
    }

    //Initiate IO
    if(setjmp(png_jmpbuf(png_ptr))) {
        perror("[write_png_gray] Error initiating IO\n");
        png_destroy_write_struct(&png_ptr, &info_ptr);
        fclose(fp);
        return 2;
    }

    png_init_io(png_ptr, fp);

    //Write header
    if(setjmp(png_jmpbuf(png_ptr))) {
        perror("[write_png_gray] Error writing header\n");
        png_destroy_write_struct(&png_ptr, &info_ptr);
        fclose(fp);
        return 2;
    }

    png_set_IHDR(png_ptr, info_ptr, width, height, bit_depth, 
            PNG_COLOR_TYPE_GRAY, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE,
            PNG_FILTER_TYPE_BASE);

    png_write_info(png_ptr, info_ptr);

    //Copy double data to 8 bit unsigned int.
    uint8_t **_image = malloc(height*sizeof(uint8_t*));
    int i,j;
    for(i = 0; i < height; i++) {
        _image[i] = malloc(width*sizeof(uint8_t));
        for(j = 0; j < width; j++) {
            _image[i][j] = (uint8_t) image[i][j];
        }
    }

    //Write data
    if(setjmp(png_jmpbuf(png_ptr))) {
        perror("[write_png_gray] Error writing bytes\n");
        png_destroy_write_struct(&png_ptr, &info_ptr);
        fclose(fp);
        for(i = 0; i < height; i++) {
            free(_image[i]);
        }
        free(_image);
        return 2;
    }

    png_write_image(png_ptr, _image);

    //Free the temporary image.
    for(i = 0; i < height; i++) {
        free(_image[i]);
    }
    free(_image);

    //End writing
    if(setjmp(png_jmpbuf(png_ptr))) {
        perror("[write_png_gray] Error writing end of file\n");
        png_destroy_write_struct(&png_ptr, &info_ptr);
        fclose(fp);
        return 2;
    }

    png_write_end(png_ptr, NULL);

    png_destroy_write_struct(&png_ptr, &info_ptr);
    fclose(fp);

    return 0;
}

/*
 * Write rgb image.
 */
int write_png_rgb(char *filename, double **red, double **green, double **blue, 
        int width, int height, int bit_depth)
{
    fp = fopen(filename, "wb");
    if(!fp) {
        return 1;
    }

    png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if(!png_ptr) {
        perror("[write_png_rgb] Error making write struct\n");
        fclose(fp);
        return 2;
    }

    info_ptr = png_create_info_struct(png_ptr);
    if(!info_ptr) {
        perror("[write_png_rgb] Error making info struct\n");
        png_destroy_write_struct(&png_ptr, NULL);
        fclose(fp);
        return 2;
    }

    if(setjmp(png_jmpbuf(png_ptr))) {
        perror("[write_png_rgb] Error initiating IO\n");
        png_destroy_write_struct(&png_ptr, &info_ptr);
        fclose(fp);
        return 2;
    }

    png_init_io(png_ptr, fp);

    if(setjmp(png_jmpbuf(png_ptr))) {
        perror("[write_png_rgb] Error writing header\n");
        png_destroy_write_struct(&png_ptr, &info_ptr);
        fclose(fp);
        return 2;
    }

    png_set_IHDR(png_ptr, info_ptr, width, height, bit_depth,
            PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE,
            PNG_FILTER_TYPE_BASE);
    png_write_info(png_ptr, info_ptr);

    uint8_t **image = malloc(height*sizeof(uint8_t*));
    int i, j;
    for(i = 0; i < height; i++) {
        image[i] = malloc(3*width*sizeof(uint8_t));
        for(j = 0; j < width; j++) {
            image[i][j*3] = (uint8_t) red[i][j];
            image[i][j*3 + 1] = (uint8_t) green[i][j];
            image[i][j*3 + 2] = (uint8_t) blue[i][j];
        }
    }

    if(setjmp(png_jmpbuf(png_ptr))) {
        perror("[write_png_rgb] Error writing data\n");
        png_destroy_write_struct(&png_ptr, &info_ptr);
        fclose(fp);
        for(i = 0; i < height; i++) {
            free(image[i]);
        }
        free(image);
        return 2;
    }

    png_write_image(png_ptr, image);

    for(i = 0; i < height; i++) {
        free(image[i]);
    }
    free(image);

    if(setjmp(png_jmpbuf(png_ptr))) {
        perror("[write_png_rgb] Error writing end of file\n");
        png_destroy_write_struct(&png_ptr, &info_ptr);
        fclose(fp);
        for(i = 0; i < height; i++) {
            free(image[i]);
        }
        free(image);
        return 2;
    }

    png_write_end(png_ptr, NULL);

    png_destroy_write_struct(&png_ptr, &info_ptr);
    fclose(fp);
    return 0;
}

/*
 * Write rgba image.
 */
int write_png_rgba(char *filename, double **red, double **green, double **blue, 
        double **alpha, int width, int height, int bit_depth)
{
    fp = fopen(filename, "wb");
    if(!fp) {
        return 1;
    }

    png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if(!png_ptr) {
        perror("[write_png_rgb] Error making write struct\n");
        fclose(fp);
        return 2;
    }

    info_ptr = png_create_info_struct(png_ptr);
    if(!info_ptr) {
        perror("[write_png_rgb] Error making info struct\n");
        png_destroy_write_struct(&png_ptr, NULL);
        fclose(fp);
        return 2;
    }

    if(setjmp(png_jmpbuf(png_ptr))) {
        perror("[write_png_rgb] Error initiating IO\n");
        png_destroy_write_struct(&png_ptr, &info_ptr);
        fclose(fp);
        return 2;
    }

    png_init_io(png_ptr, fp);

    if(setjmp(png_jmpbuf(png_ptr))) {
        perror("[write_png_rgb] Error writing header\n");
        png_destroy_write_struct(&png_ptr, &info_ptr);
        fclose(fp);
        return 2;
    }

    png_set_IHDR(png_ptr, info_ptr, width, height, bit_depth,
            PNG_COLOR_TYPE_RGBA, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE,
            PNG_FILTER_TYPE_BASE);
    png_write_info(png_ptr, info_ptr);

    uint8_t **image = malloc(height*sizeof(uint8_t*));
    int i, j;
    for(i = 0; i < height; i++) {
        image[i] = malloc(4*width*sizeof(uint8_t));
        for(j = 0; j < width; j++) {
            image[i][j*4] = (uint8_t) red[i][j];
            image[i][j*4 + 1] = (uint8_t) green[i][j];
            image[i][j*4 + 2] = (uint8_t) blue[i][j];
            image[i][j*4 + 3] = (uint8_t) alpha[i][j];
        }
    }

    if(setjmp(png_jmpbuf(png_ptr))) {
        perror("[write_png_rgb] Error writing data\n");
        png_destroy_write_struct(&png_ptr, &info_ptr);
        fclose(fp);
        for(i = 0; i < height; i++) {
            free(image[i]);
        }
        free(image);
        return 2;
    }

    png_write_image(png_ptr, image);

    for(i = 0; i < height; i++) {
        free(image[i]);
    }
    free(image);

    if(setjmp(png_jmpbuf(png_ptr))) {
        perror("[write_png_rgb] Error writing end of file\n");
        png_destroy_write_struct(&png_ptr, &info_ptr);
        fclose(fp);
        for(i = 0; i < height; i++) {
            free(image[i]);
        }
        free(image);
        return 2;
    }

    png_write_end(png_ptr, NULL);

    png_destroy_write_struct(&png_ptr, &info_ptr);
    fclose(fp);
    return 0;
}
