/*
 * This file is a header file for easy reading and writing of grayscale png 
 * images. Always check return values. This file is free to be used and 
 * modified in any way one sees fit.
 */
#ifndef READ_PNG_H
#define READ_PNG_H

#ifdef __cplusplus
extern "C" {
#endif

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <png.h>

#define HEADER_SIZE 8

FILE *fp;
png_structp png_ptr;
png_infop info_ptr, stop_info;
int color;

/*
 * Start reading the PNG file. Get height, width and bit depth of the image.
 * Have to read this before reading the image itself. 
 * 
 * Returns:
 *      0 - Success.
 *      1 - File not found.
 *      2 - Not a PNG file according to header.
 *      3 - All else.
 */
int read_png_info(char *filename, int *width, int *height, int *bit_depth);

/*
 * Read a grayscale image as double. read_png_info have to be called before this
 * function, and have exited successfully.
 *
 * Returns:
 *      0 - Success.
 *      1 - Function not ready to be run. (Missing read_png_info).
 *      2 - Not a grayscale image.
 *      3 - Unexpected error occurred during run.
 */
int read_png_gray(double **image, int width, int height);

/*
 * Read a RGB image as doubles. read_png_info have to be called before this
 * function, found that it's an RGB image and exited successfully.
 *
 * Returns:
 *      0 - Success.
 *      1 - Function not ready to be run. (Missing read_png_info).
 *      2 - Not an RGB image.
 *      3 - Unexpected error occured during run.
 */
int read_png_rgb(double **red, double **green, double **blue, int width, 
        int height);

/*
 * Read a RGBA image as doubles. read_png_info have to be called before this 
 * function, found that it's an RGBA image and exited successfully.
 *
 * Returns:
 *      0 - Success.
 *      1 - Function not ready to be run. (Missing read_png_info).
 *      2 - Not an RGBA image.
 *      3 - Unexpected error occured during run.
 */
int read_png_rgba(double **red, double **green, double **blue, double **alpha,
        int width, int height);

/*
 * Write grayscale data to a PNG file. Need width, height and bit depth as 
 * parameters. If you run this between read_png_info and read_png_*, you 
 * can no longer call read_png_*.
 *
 * Returns:
 *      0 - Success.
 *      1 - Error opening file for writing.
 *      2 - Any unexpected errors during write.
 */
int write_png_gray(char *filename, double **image, int width, int height, 
        int bit_depth);

/*
 * Write RGB data to a PNG file. Need width, height and bit depth as 
 * parameters. If you run this between read_png_info and read_png_*, you can
 * no longer call read_png_*.
 *
 * Returns:
 *      0 - Success.
 *      1 - Error opening file for writing.
 *      2 - Any unexpected errors during write.
 */
int write_png_rgb(char *filename, double **red, double **green, double **blue,
        int width, int height, int bit_depth);

/*
 * Write RGBA data to a PNG file. Need width, height and bit depth as 
 * parameters. If you run this between read_png_info and read_png_*, you can
 * no longer call read_png_*.
 *
 * Returns:
 *      0 - Success.
 *      1 - Error opening file for writing.
 *      2 - Any unexpected errors during write.
 */
int write_png_rgba(char *filename, double **red, double **green, double **blue,
        double **alpha, int width, int height, int bit_depth);

#ifdef __cplusplus
}
#endif
#endif
